import threading

from django.conf import settings
from django.core.cache import caches
from django.core.exceptions import ImproperlyConfigured
from large_image.cache_util.base import BaseCache
from large_image.exceptions import TileCacheConfigurationError


class DjangoCache(BaseCache):
    """Use Django cache as the backing cache for large-image."""

    def __init__(self, cache, alias, getsizeof=None):
        super().__init__(0, getsizeof=getsizeof)
        self._django_cache = cache
        self._alias = alias

    def __repr__(self):  # pragma: no cover
        return f'DjangoCache<{repr(self._alias)}>'

    def __iter__(self):  # pragma: no cover
        # return invalid iter
        return None

    def __len__(self):  # pragma: no cover
        # return invalid length
        return -1

    def __contains__(self, key):
        hashed_key = self._hashKey(key)
        return self._django_cache.__contains__(hashed_key)

    def __delitem__(self, key):
        hashed_key = self._hashKey(key)
        return self._django_cache.delete(hashed_key)

    def __getitem__(self, key):
        hashed_key = self._hashKey(key)
        value = self._django_cache.get(hashed_key)
        if value is None:
            return self.__missing__(key)
        return value

    def __setitem__(self, key, value):
        hashed_key = self._hashKey(key)
        # TODO: do we want to use `add` instead to add a key only if it doesn’t already exist
        return self._django_cache.set(hashed_key, value)

    @property
    def curritems(self):  # pragma: no cover
        raise NotImplementedError

    @property
    def currsize(self):  # pragma: no cover
        raise NotImplementedError

    @property
    def maxsize(self):  # pragma: no cover
        raise NotImplementedError

    def clear(self):
        self._django_cache.clear()

    @staticmethod
    def getCache():  # noqa: N802
        try:
            name = getattr(settings, 'LARGE_IMAGE_CACHE_NAME', 'default')
            django_cache = caches[name]
        except ImproperlyConfigured:  # pragma: no cover
            raise TileCacheConfigurationError
        cache_lock = threading.Lock()
        cache = DjangoCache(django_cache, alias=name)
        return cache, cache_lock



from django.http import HttpResponse
from drf_spectacular.utils import extend_schema
from large_image.exceptions import TileSourceXYZRangeError
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.request import Request
from rest_framework.response import Response

from django_large_image import tilesource
from django_large_image.rest import params
from django_large_image.rest.base import LargeImageMixinBase
from django_large_image.rest.renderers import image_renderers
from django_large_image.rest.serializers import TileMetadataSerializer

tile_metadata_summary = 'Returns tile metadata.'
tile_metadata_parameters = params.BASE
tile_summary = 'Returns tile image binary.'
tile_parameters = params.BASE + [params.z, params.x, params.y, params.fmt_png] + params.STYLE
tile_corners_summary = 'Returns bounds of a tile for a given x, y, z index.'
tile_corners_parameters = params.BASE + [params.z, params.x, params.y]


class TilesMixin(LargeImageMixinBase):
    @extend_schema(
        methods=['GET'],
        summary=tile_metadata_summary,
        parameters=tile_metadata_parameters,
    )
    @action(detail=False, url_path=r'tiles/metadata')
    def tiles_metadata(self, request: Request, pk: int = None, **kwargs) -> Response:
        source = self.get_tile_source(request, pk, style=False)
        source.dli_geospatial = tilesource.is_geospatial(source)
        serializer = TileMetadataSerializer(source)
        return Response(serializer.data)

    @extend_schema(
        methods=['GET'],
        summary=tile_summary,
        parameters=tile_parameters,
    )
    @action(
        detail=False,
        url_path=rf'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+).{params.FORMAT_URL_PATTERN}',
        renderer_classes=image_renderers,
    )
    def tile(
        self, request: Request, x: int, y: int, z: int, pk: int = None, fmt: str = 'png', **kwargs
    ) -> HttpResponse:
        encoding = tilesource.format_to_encoding(fmt, pil_safe=True)
        source = self.get_tile_source(request, pk, encoding=encoding)
        style = self.get_style(request) or {}
        frame = request.query_params.get('frame') or style.get('frame')
        try:
            tile_binary = source.getTile(int(x), int(y), int(z), frame=frame)
        except TileSourceXYZRangeError as e:
            raise ValidationError(e)
        mime_type = source.getTileMimeType()
        return HttpResponse(tile_binary, content_type=mime_type)

    @extend_schema(
        methods=['GET'],
        summary=tile_corners_summary,
        parameters=tile_corners_parameters,
    )
    @action(
        detail=False, methods=['get'], url_path=r'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+)/corners'
    )
    def tile_corners(
        self, request: Request, x: int, y: int, z: int, pk: int = None, **kwargs
    ) -> HttpResponse:
        source = self.get_tile_source(request, pk, style=False)
        xmin, ymin, xmax, ymax = source.getTileCorners(int(z), int(x), int(y))
        metadata = {
            'xmin': xmin,
            'xmax': xmax,
            'ymin': ymin,
            'ymax': ymax,
            # 'proj4': source.getProj4String(),  # not supported by rasterio
        }
        return Response(metadata)


class TilesDetailMixin(TilesMixin):
    @extend_schema(
        methods=['GET'],
        summary=tile_metadata_summary,
        parameters=tile_metadata_parameters,
    )
    @action(detail=True, url_path=r'tiles/metadata')
    def tiles_metadata(self, request: Request, pk: int = None, **kwargs) -> Response:
        return super().tiles_metadata(request, pk)

    @extend_schema(
        methods=['GET'],
        summary=tile_summary,
        parameters=tile_parameters,
    )
    @action(
        detail=True,
        url_path=rf'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+).{params.FORMAT_URL_PATTERN}',
        renderer_classes=image_renderers,
    )
    def tile(
        self, request: Request, x: int, y: int, z: int, pk: int = None, fmt: str = 'png', **kwargs
    ) -> HttpResponse:
        return super().tile(request, x, y, z, pk, fmt)

    @extend_schema(
        methods=['GET'],
        summary=tile_corners_summary,
        parameters=tile_corners_parameters,
    )
    @action(
        detail=True, methods=['get'], url_path=r'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+)/corners'
    )
    def tile_corners(
        self, request: Request, x: int, y: int, z: int, pk: int = None, **kwargs
    ) -> HttpResponse:
        return super().tile_corners(request, x, y, z, pk)



from rest_framework.renderers import BaseRenderer as RFBaseRenderer


class BaseRenderer(RFBaseRenderer):
    render_style = 'binary'
    charset = None

    def render(self, data, media_type=None, renderer_context=None):
        return data


class PNGRenderer(BaseRenderer):
    media_type = 'image/png'
    format = 'png'


class JPEGRenderer(BaseRenderer):
    media_type = 'image/jpeg'
    format = 'jpeg'


class JPGRenderer(BaseRenderer):
    media_type = 'image/jpeg'
    format = 'jpg'


image_renderers = [PNGRenderer, JPEGRenderer, JPGRenderer]


class TifRenderer(BaseRenderer):
    media_type = 'image/tiff'
    format = 'tif'


class TiffRenderer(BaseRenderer):
    media_type = 'image/tiff'
    format = 'tiff'


image_data_renderers = image_renderers + [TifRenderer, TiffRenderer]



# flake8: noqa: F401
from django_large_image.rest.standalone import (
    ListAvailableFormatsView,
    ListColormapsView,
    ListTileSourcesView,
)
from django_large_image.rest.viewsets import (
    LargeImageDetailMixin,
    LargeImageFileDetailMixin,
    LargeImageMixin,
    LargeImageVSIFileDetailMixin,
)



import logging

from django.apps import AppConfig
from django.conf import settings
import large_image

logger = logging.getLogger(__name__)


class DjangoLargeImageConfig(AppConfig):
    name = 'django_large_image'
    verbose_name = 'Django Large Image'
    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self):
        # Set up cache with large_image
        # Use any existing settings defined by other installed apps,
        # Defaulting to caching with django
        large_image.config.setConfig(
            'cache_backend',
            getattr(settings, 'LARGE_IMAGE_CACHE_BACKEND', 'django'),
        )

        # Set any other cache config if settings are defined
        for config_var_name in [
            'cache_python_memory_portion',
            'cache_memcached_url',
            'cache_memcached_username',
            'cache_memcached_password',
            'cache_redis_url',
            'cache_redis_username',
            'cache_redis_password',
            'cache_tilesource_memory_portion',
            'cache_tilesource_maximum',
            'cache_sources',
        ]:
            settings_attr_name = 'LARGE_IMAGE_' + config_var_name.upper()
            settings_attr_value = getattr(settings, settings_attr_name, None)
            if settings_attr_value is not None:
                large_image.config.setConfig(config_var_name, settings_attr_value)
